/*  =================================================== */
/*  Intelligent FRAC.		(C) 2000  Michael Glickman      */
/*  --------------------------------------------------- */
/*  See LICENSE regarding distribution policy and       */
/*  conditions of use.                                  */ 
/*  =================================================== */

#include <stdio.h>
#include <vgakeyboard.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <ctype.h>
#include "ifr.h"

#if defined(HAVE_SYS_KD_H)
#include <sys/kd.h>
#elif defined(HAVE_LINUX_KD_H)
#include <linux/kd.h>
#endif

extern char  *board;
typedef struct 
{ int key1; int key2; ACTION act; } SRKEYS;

static  int  fd;
static  int  capslock;
static  int  leds;

extern  GAME game;
extern  int	HelpMode;
extern  int	kbd_raw;
extern SRKEYS PlayerKeys[], ShiftKeys[], CommonKeys[], EditKeys[];
extern const int PlayerKeysCount, ShiftKeysCount, CommonKeysCount, EditKeysCount;
extern const unsigned char *AsciiTables[];
extern const int  AsciTableSizes[];

#ifdef USE_VOLUME_CONTROL
extern SRKEYS VolumeKeys[];
extern const int VolumeKeysCount;
#endif

static void kbd_get_leds(void)
{
	leds = 0;
#if defined(KDGETLED)
	if (fd >=0) ioctl(fd, KDGETLED, &leds);
#endif  	
	return;
}

static void kbd_toggle_capslock(void)
{
#if defined(KDSETLED)
	if (fd >=0)
	{ int stat = leds ^ LED_CAP;
		if (ioctl(fd, KDSETLED, stat) >=0) leds = stat;
	} 
#endif
	return;
}	

int kbd_interface_start(void)
{	
	kbd_raw = 1;
	capslock = 0;
	fd = keyboard_init_return_fd();
	if (fd >= 0)
	{	kbd_get_leds();
		keyboard_translatekeys(TRANSLATE_CURSORKEYS | TRANSLATE_KEYPADENTER);
		return 1;
	}
	return 0;
}

void kbd_interface_end(void)
{
#if defined(KDSETLED)
	if (fd >=0) ioctl(fd, KDSETLED, -1);
#endif  	
		keyboard_close();
}


static ACTION get_kbd_action(char *buf, SRKEYS table[], int tabsize)
{		int i, key;
		
		for (i=0; i<tabsize; i++)
		{  if (buf[table[i].key1] ||
           ((key=table[i].key2) != 0 && buf[key])
			    )
					return  table[i].act; 
		}
		return ACT_NONE;
}

static ACTION process_kbd_buffer(PLAYER *player, char *buf)
{	int player_type = player && player->type;	
	int shift_state;
	ACTION act = ACT_NONE;

  if (HelpMode || game.level == SummaryLevel)
		shift_state = 1;
	else
		shift_state = buf[SCANCODE_LEFTSHIFT] | buf[SCANCODE_RIGHTSHIFT] |
                  buf[SCANCODE_LEFTALT] | buf[SCANCODE_RIGHTALT] ;	

#ifdef USE_VOLUME_CONTROL
  if (shift_state || (player_type > 0 && game.level < Levels))
	  act = get_kbd_action(buf, VolumeKeys, VolumeKeysCount);

  if (act == ACT_NONE && player_type == 0)
#else
  if (player_type == 0)
#endif
	{	
		if (shift_state)
	 		act = get_kbd_action(buf, ShiftKeys, ShiftKeysCount);
		else
		 	act = get_kbd_action(buf, PlayerKeys, PlayerKeysCount);
	}

	if (act == ACT_NONE) 
	{	act = get_kbd_action(buf, CommonKeys, CommonKeysCount);
		if(shift_state && act == ACT_ROTATEFWD) act = ACT_ROTATEBACK;
	}

	if  ((act == ACT_TAB || act == ACT_DROP) && shift_state)
			  act = ACT_BACKTAB;

	return act;
}

static int lookup_ascii(char *buf)
{	const unsigned char *tabptr;
	char  chr = 0;
	int i, tabptr_size, index = 0;

	if (buf[SCANCODE_LEFTSHIFT] | buf[SCANCODE_RIGHTSHIFT])
			index++;

#ifdef LED_CAP
	if (leds & LED_CAP) index += 2;
#endif

	tabptr = AsciiTables[index];
	tabptr_size = AsciTableSizes[index];
	

	for(i=0; i< tabptr_size; i++)
	{	chr = *tabptr++;
		if (chr && buf[i]) break;
	}
	
	if (i >= tabptr_size) return 0;
	return 0x100 | chr;	
}


static ACTION process_kbd_buffer_edit(PLAYER *player, char *buf)
{		ACTION act;

    act = get_kbd_action(buf, EditKeys, EditKeysCount);

		if (act == ACT_NONE)
			act = (ACTION) lookup_ascii(buf);

		return act;

}



ACTION kbd_interface_routine(PLAYER *player, int edit_mode)
{	char *buf;
	int  capslock_new;	
	
  if (keyboard_update() == 0) return ACT_NONE;
	
	buf = keyboard_getstate();
	
	if (buf == NULL) return ACT_NONE;

	capslock_new = buf [SCANCODE_CAPSLOCK];
	if (capslock_new != capslock)
	{	 capslock = capslock_new;
		 if (capslock)  kbd_toggle_capslock();
	}

	return edit_mode ? process_kbd_buffer_edit(player, buf) :
                     process_kbd_buffer(player, buf);
}

void flush_kbd(void)
{
	return;
}

