/*  =================================================== */
/*  Intelligent FRAC.		(C) 2000  Michael Glickman      */
/*  --------------------------------------------------- */
/*  See LICENSE regarding distribution policy and       */
/*  conditions of use.                                  */ 
/*  =================================================== */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <ctype.h>

#include "ifr.h"

extern Display	*disp;
extern Window	wnd;
extern char  *board;
extern CMAP_TYPE cm_type;
extern GAME  game;

#ifndef USE_XF86DGA
static Atom    WMProtAtom, WMDelAtom;
#endif

void visualize_pane(PANE_TYPE which_pane, int x, int y, int wid, int hei);

typedef struct 
{ KeySym key1; KeySym key2; KeySym key3; ACTION act; } XKEYS;

extern  GAME game;
extern  int	HelpMode;
extern XKEYS PlayerKeys[], ShiftKeys[], CommonKeys[], EditKeys[];
extern const int PlayerKeysCount, ShiftKeysCount,  CommonKeysCount, EditKeysCount;

#ifdef USE_VOLUME_CONTROL
extern XKEYS VolumeKeys[];
extern const int VolumeKeysCount;
#endif

int kbd_interface_start(void)
{
    XSelectInput(disp, wnd, ExposureMask | KeyPressMask 
	         | StructureNotifyMask | SubstructureNotifyMask | VisibilityChangeMask);

#ifdef USE_XF86DGA
    XGrabKeyboard(disp, wnd, False, GrabModeAsync, GrabModeAsync, CurrentTime);
#else
    WMProtAtom = XInternAtom(disp, "WM_PROTOCOLS", True);
    WMDelAtom = XInternAtom(disp, "WM_DELETE_WINDOW", True);
    
    if (WMProtAtom != (Atom)None && WMDelAtom != (Atom)None)
		XSetWMProtocols(disp, wnd, &WMDelAtom, 1);
    XMapWindow(disp, wnd);
#endif

	//XFlush(disp);    
		flush_all_pane(PT_FOREGROUND);
     

#ifndef USE_XF86DGA
//   if (cm_type == CM_PALETTE)
   {    XWarpPointer(disp, None, wnd, 0, 0, 0, 0, 10, 10);
/*      XSetInputFocus(disp, wnd, RevertToParent, CurrentTime);   */
   }    
#endif

   return 1;
}

void kbd_interface_end(void)
{
#ifdef USE_XF86DGA
    XUngrabKeyboard(disp, CurrentTime);
#else
    XUnmapWindow(disp, wnd);
#endif
}

#ifndef USE_XF86DGA
static int process_nonkey_event(PLAYER *player, XEvent *evnt)
{
    if (XFilterEvent(evnt, wnd)) return 0;

    switch (evnt->type)
    {
	case  Expose:
	{
	    XExposeEvent *xee;
	    game.status &= ~GMST_SUSPENDED;
	    xee = (XExposeEvent *) evnt;
//		if (xee->count == 0)
	    flush_pane(PT_FOREGROUND, 
		  xee->x, xee->y, xee->width, xee->height);
	}	    
	break;
    
	case MapNotify:
	    game.status &= ~GMST_SUSPENDED;
	    break;
		
	case UnmapNotify:
	if (game.level < Levels && player->type == 0)
			game.status |= GMST_SUSPENDED;
	    break;
	
	case ClientMessage:
	{
	    XClientMessageEvent *xcme;
	    xcme = (XClientMessageEvent *)evnt;    
	    if (xcme->message_type == WMProtAtom &&	
    		   xcme->data.l[0] == WMDelAtom)	
	    return 1;
	}
	break;	       		       	

    }

    return 0;
}
#endif

static ACTION get_key_action(KeySym ks, XKEYS table[], int tabsize)
{		int i;

	for (i=0; i<tabsize; i++)
		if (ks == table[i].key1 ||
    	  ks == table[i].key2 ||  
        ks == table[i].key3)
				  return table[i].act; 

	return ACT_NONE;
}

static ACTION process_key_event(PLAYER *player, XEvent *evnt)
{
  KeySym ks;
  ACTION act = ACT_NONE;
  int player_type = player && player->type;	
  int shift_state;
    
  if (XFilterEvent(evnt, wnd)) return ACT_NONE ;

  ks = XLookupKeysym((XKeyEvent *)evnt, 0);
/*    printf  ("KeySym = %lx\n", ks); */
     
  if (ks >= XK_a && ks <= XK_z)
	ks -= (XK_a - XK_A);		
  else
  if (ks >= XK_KP_Home && ks <= XK_KP_Begin)
	 		ks -= (XK_KP_Home - XK_Home);		

  if (ks >= XK_KP_0 && ks <= XK_KP_9)
	  ks -= (XK_KP_0 - XK_0);		
  else
	  if (ks == 0) ks = XK_Begin;		


  if (HelpMode || game.level == SummaryLevel)
	shift_state = 1;
  else
	shift_state = (evnt->xkey).state & (ShiftMask | Mod1Mask);	

#ifdef USE_VOLUME_CONTROL
  if (shift_state || (player_type > 0 && game.level < Levels))
	  act = get_key_action(ks, VolumeKeys, VolumeKeysCount);

  if (act == ACT_NONE && player_type == 0)
#else
  if (player_type == 0)
#endif
  { if (shift_state)
	  act = get_key_action(ks, ShiftKeys, ShiftKeysCount);
	else      
	  act = get_key_action(ks, PlayerKeys, PlayerKeysCount);
  }	  

  if (act == ACT_NONE)
  {  act = get_key_action(ks, CommonKeys, CommonKeysCount);
		if(shift_state && act == ACT_ROTATEFWD) act = ACT_ROTATEBACK; 
  }

  if  ((act == ACT_TAB || act == ACT_DROP) && shift_state)
							 act = ACT_BACKTAB;

  return act;
}

static ACTION process_key_event_edit(PLAYER *player, XEvent *evnt)
{
    KeySym ks;
	char buffer[2];
	int  c;  

    if (XFilterEvent(evnt, wnd)) return ACT_NONE;
	
	c = XLookupString((XKeyEvent *)evnt, buffer, 2, &ks, 0);
	if (c == 1)
	{  c = buffer[0];
//		if (!isprint(c))
       if (!iscntrl(c)) return (ACTION) (0x100 | c);
	   
	}
			 
	 return  get_key_action(ks, EditKeys, EditKeysCount);

}

ACTION kbd_interface_routine(PLAYER *player, int edit_mode)
{	
    XEvent  evnt;
    ACTION  act = ACT_NONE;
    int    evnt_count;

/*	evnt_count = XPending(disp);  */
    evnt_count = XEventsQueued(disp, QueuedAfterReading);

    while (--evnt_count >= 0)
    {   XNextEvent(disp, &evnt);
	
	if (evnt.type == KeyPress)
	{  if (act == ACT_NONE)
           act = edit_mode ? process_key_event_edit(player, &evnt)
				                : process_key_event(player, &evnt);
//		   else    	      
//	    	 XPutBackEvent(disp, &evnt); 
	}
#ifndef USE_XF86DGA
	else
	if (process_nonkey_event(player, &evnt))
			{  act = ACT_QUIT; 
			   break;
			}	   
	
#endif
  }
    
  return act;
}

void flush_kbd(void)
{
	XEvent evnt;
	while (XCheckTypedEvent(disp, KeyPress, &evnt) == True);

}
